/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Sun Feb 24 16:11:00 1991 by burrows        */
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include "indexserver.h"

static struct msgset *doop ();
static struct msgset *wordmulti ();
static struct msgset *wordmentions ();
static struct msgset *allfunc ();
static struct msgset *nullfunc ();
static struct msgset *earliestfunc ();
static struct msgset *latestfunc ();
static struct msgset *onfunc ();
static struct msgset *afterfunc ();
static struct msgset *beforefunc ();
static struct msgset *pathnamefunc ();
static void unexpected ();
static void unexpectedstr ();
static void unknown ();

int nc (istring)
register char **istring;
{
	register int c = **istring;

	if (c != '\0')
		(*istring)++;
	return (c);
}

struct msgset *getmsgset (index2, groups, string, perr)
struct index2 *index2;
struct groups *groups;
char *string;
int *perr;
{
	struct msgset *current;
	int c;
	
	c = nc (&string);
	current = submsgset (index2, groups, index2->allmsgs, index2->allmsgs,
		nc, &string, &c, 0, perr);

	if (interrupt) {
		(void) printf ("?interrupt\n");
		*perr = 1;
	}

	if (*perr) {
		freemsgset (current);
		return (0);
	}

	return (current);
}

struct msgset *submsgset (index2, groups, fulldomain, domain,
	nextc, string, pc, level, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
int (*nextc) ();
register char **string;
int *pc;
int level;
int *perr;
{
	int c = *pc;
	struct msgset *current = 0, *x, *tmp, *subdomain;
	int not;
	int or = 1;
	char buf[256], arg[512];
	int i, j;

	while (isspace (c) && (c != '\n' || level == 0))
		c = (*nextc) (string);
	
	while (c != '\0' && c != ')' && c != '\n' && !interrupt) {
		not = 0;
		subdomain = (or? domain: current);

notagain:
		while ((isspace (c) && (c != '\n' || level == 0)) || c == '!') {
			if (c == '!')
				not = !not;
			c = (*nextc) (string);
		}

		if (c == '(') {
			c = (*nextc) (string);
			x = submsgset (index2, groups, fulldomain, subdomain,
				nextc, string, &c, level+1, perr);
			if (*perr != 0) {
				freemsgset (current);
				freemsgset (x);
				return (0);
			}
			if (c == ')')
				c = (*nextc) (string);
			else if (c != '\0') {
				freemsgset (current);
				freemsgset (x);
				unexpected (c, "')'");
				*perr = 1;
				return (0);
			}
		} else if (isalpha (c) || c == '@' || c == '\'' || c == '_') {
			i = 0;

			do {
				if (i < sizeof (buf) - 1)
					buf[i++] = c;
				c = (*nextc) (string);
			} while (isalnum (c) || c == '_' || c=='@' || c=='\'');
			buf[i] = '\0';

			if (strcmp (buf, "not") == 0) {
				not = !not;
				goto notagain;
			}

			while (isspace (c) && (c != '\n' || level == 0))
				c = (*nextc) (string);
			
			j = -1;
			if (c == '(') {
				int nest = 0;

				j = 0;
				c = (*nextc) (string);
				while (c != '\0' && (c != ')' || nest > 0)) {
					if (c == '(')
						nest++;
					else if (c == ')' && nest > 0)
						nest--;
					else if (c == '\\') {
						c = (*nextc) (string);
						if (c == '\0')
							break;
					}
					if (j < sizeof (arg) - 1)
						arg[j++] = c;
					c = (*nextc) (string);
				}
				arg[j] = '\0';

				if (c == ')')
					c = (*nextc) (string);
			}

			x = doop (index2, groups, fulldomain, subdomain, buf, i,
				(j>=0?arg: (char *)0), j, not, perr);
			if (*perr != 0) {
				freemsgset (current);
				freemsgset (x);
				return (0);
			}
			not = 0;
		} else {
			unexpected (c, "'(' or function name");
			*perr = 1;
			freemsgset (current);
			return (0);
		}

		if (not) {
			tmp = notmsgset (subdomain, x);
			freemsgset (x);
			x = tmp;
		}

		if (or && current != 0)
			x = into_ormsgset (x, current);

		freemsgset (current);
		current = x;
			
		while (isspace (c) && (c != '\n' || level == 0))
			c = (*nextc) (string);
		
		if (c == '&' || c == '|') {
			or = (c == '|');

			c = (*nextc) (string);

			if ((c == '&' && !or) || (c == '|' && or))
				c = (*nextc) (string);

			while (isspace (c))
				c = (*nextc) (string);
		} else if (isalpha (c)) {
			i = 0;

			do {
				if (i < sizeof (buf) - 1)
					buf[i++] = c;
				c = (*nextc) (string);
			} while (isalnum (c) || c == '_' || c=='@' || c=='\'');
			buf[i] = '\0';

			if (strcmp (buf, "and") == 0)
				or = 0;
			else if (strcmp (buf, "or") == 0)
				or = 1;
			else {
				unexpectedstr (buf, "and/or");
				*perr = 1;
				freemsgset (current);
				return (0);
			}

			while (isspace (c))
				c = (*nextc) (string);
		}
	}

	*pc = c;

	return (current);
}

/* ARGSUSED */
static struct msgset *nullfunc (index2, groups, fulldomain, domain,
	arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *arg;
int arglen;
int not;
int *perr;
{
	return (0);
}

/* ARGSUSED */
static struct msgset *allfunc (index2, groups, fulldomain, domain,
	arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *arg;
int arglen;
int not;
int *perr;
{
	if (not)
		return (0);
	else
		return (ormsgset (domain, (struct msgset *) 0));
}

struct op {
	char tc;
	char *op;
	struct msgset *(*func) ();
	int (*selectfunc) ();
	char *selectarg;
	int expectarg;		/* 1 = string arg, 2 = msgset arg */
};

struct op ops[] = {

	{ 'A',	"All",		allfunc,    nullselect,	(char *) 0,	0},
	{ 'F',	"From_to_cc",	nullfunc,   fromtoccselect,(char *) 0,	1},
	{ 'I',	"Id",		nullfunc,   idselect,	(char *) 0,	1},
	{ 'T',	"To_cc",	nullfunc,   toccselect,	(char *) 0,	1},
	{ 'a',	"after",	afterfunc,  nullselect,	(char *) 0,	1},
	{ 'b',	"before",	beforefunc, nullselect,	(char *) 0,	1},
	{ 'c',	"cc",		nullfunc,   ccselect,	(char *) 0,	1},
	{ 'd',	"discussion",	discussfunc,nullselect, (char *) 0,	2},
	{ 'e',	"earliest",	earliestfunc,nullselect, (char *) 0,	2},
	{ 'f',	"from",		nullfunc,   fromselect,	(char *) 0,	1},
	{ 'g',	"group",	nullfunc,   groupselect,(char *) 0,	1},
	{ 'i',	"inreplyto",	nullfunc,   repselect,	(char *) 0,	1},
	{ 'k',	"keyword",	nullfunc,   keyselect,	(char *) 0,	1},
	{ 'l',	"latest",	latestfunc, nullselect, (char *) 0,	2},
	{ 'm',	"messageid",	nullfunc,   midselect,	(char *) 0,	1},
	{ 'o',	"on",		onfunc,     nullselect,	(char *) 0,	1},
	{ 'p',	"pathname",	pathnamefunc, nullselect, (char *) 0,	1},
	{ 'r',	"references",	nullfunc,   refselect,	(char *) 0,	1},
	{ 's',	"subject",	nullfunc,   subjselect,	(char *) 0,	1},
	{ 't',	"to",		nullfunc,   toselect,	(char *) 0,	1},
	{ 'w',	"word",		wordmentions,nullselect,(char *) 0,	1},
	{ 'W',  "Words",	wordmulti,  nullselect, (char *) 0,	1},

	{ 0,   (char *) 0 }
};

/* ARGSUSED */
static struct msgset *doop (index2, groups, fulldomain, domain,
	op, oplen, arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *op;
int oplen;
char *arg;
int arglen;
int not;
int *perr;
{
	register int i;
	char var[128];
	struct msgset *msgset;
	int found;

	if (op[0] == '@' || op[0] == '\'') {
		if (not)
			return (notmsgset (domain, getvar (op+1, &found)));
		else
			return (andmsgset (domain, getvar (op+1, &found)));
	}

	for (i = 0; ops[i].op != 0; i++) {
		if (strncmp (ops[i].op, op, oplen) == 0) {
			if (ops[i].expectarg && arg == 0) {
				unexpectedstr ("no argument", "(arg)");
				*perr = 1;
				return (0);
			} else if (!ops[i].expectarg && arg != 0) {
				unexpectedstr ("argument", "no argument");
				*perr = 1;
				return (0);
			}

			var[0] = '\0';
			if (fulldomain == domain &&
			    domain == index2->allmsgs && arg != 0 &&
			    arglen < sizeof (var) - 30) {
				(void) sprintf (var, "%s,%d,%s", ops[i].op,
					not?1:0, arg);
				found = 0;
				msgset = getvar (var, &found);
				if (found) {
/* (void) fprintf (stderr, "%s cache hit on %s\n", ops[i].op, var); */
					return (ormsgset (msgset,
						(struct msgset *) 0));
				}
/* else (void) fprintf (stderr, "%s cache miss on %s\n", ops[i].op, var); */
			}

			if (ops[i].func != nullfunc) {
				msgset = (*ops[i].func) (index2, groups,
				    fulldomain, domain, arg, arglen, not, perr);
			} else if (ops[i].selectarg != 0) {
				msgset = filterselectmsg (index2, domain,
				   ops[i].selectfunc, ops[i].selectarg, not, 0);
			} else
				msgset = filterselectmsg (index2, domain,
				   ops[i].selectfunc, arg, not, 0);
			
			if (var[0] != '\0' && !*perr)
				setvar (var, msgset);

			return (msgset);
		}
	}

	unknown (op);
	*perr = 1;

	return (0);
}

static void unexpected (c, ec)
char c;
char *ec;
{
	char got[4];

	got[0] = '\''; got[1] = c; got[2] = '\''; got[3] = '\0';
	unexpectedstr (got, ec);
}

static void unexpectedstr (got, ec)
char *got;
char *ec;
{
	(void) printf ("?expected %s, found %s\n", ec, got);
}

static void unknown (s)
char *s;
{
	(void) printf ("?unknown selector '%s'\n", s);
}

/* ARGSUSED */
static struct msgset *earliestfunc (index2, groups, fulldomain, domain,
	arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *arg;
int arglen;
int not;
int *perr;
{
	int c;
	struct msgset *tmp, *result;
	register long *x, l;
	register int i, j, k;
	long best = -1;
	char *beststr = "ffffff";
	char *f;

	c = nc (&arg);
	tmp = submsgset (index2, groups, fulldomain, fulldomain,
		nc, &arg, &c, 1, perr);

	if (tmp == 0)
		return (tmp);

	x = tmp->bitset;
	for (i = 0, j = tmp->nbitset; i < j; i++) {
		if ((l = x[i]) != 0) {
			for (k = 0; k < LONGBITS; k++) {
				if (l & (1 << k)) {
					f = field (index2, (long) k+i*LONGBITS,
						DATEFIELD);
					if (strcmp (f, beststr) < 0) {
						best = k+i*LONGBITS;
						beststr = f;
					}
				}
			}
		}
	}

	if (best < 0)
		return (tmp);

	if (not) {
		x[ROUNDDOWN (best)] &= ~(1 << BIT (best));
		result = tmp;
	} else {
		result = ormsgset (tmp, (struct msgset *) 0);
		x[ROUNDDOWN (best)] &= ~(1 << BIT (best));
		result = into_notmsgset (result, tmp);
		freemsgset (tmp);
	}

	result = into_andmsgset (result, domain);

	return (result);
}

/* ARGSUSED */
static struct msgset *latestfunc (index2, groups, fulldomain, domain,
	arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *arg;
int arglen;
int not;
int *perr;
{
	int c;
	struct msgset *tmp, *result;
	register long *x, l;
	register int i, j, k;
	long best = -1;
	char *beststr = "000000";
	char *f;

	c = nc (&arg);
	tmp = submsgset (index2, groups, fulldomain, fulldomain,
		nc, &arg, &c, 1, perr);

	if (tmp == 0)
		return (tmp);

	x = tmp->bitset;
	for (i = 0, j = tmp->nbitset; i < j; i++) {
		if ((l = x[i]) != 0) {
			for (k = 0; k < LONGBITS; k++) {
				if (l & (1 << k)) {
					f = field (index2, (long) k+i*LONGBITS,
						DATEFIELD);
					if (strcmp (f, beststr) >= 0) {
						best = k+i*LONGBITS;
						beststr = f;
					}
				}
			}
		}
	}

	if (best < 0)
		return (tmp);

	if (not) {
		x[ROUNDDOWN (best)] &= ~(1 << BIT (best));
		result = tmp;
	} else {
		result = ormsgset (tmp, (struct msgset *) 0);
		x[ROUNDDOWN (best)] &= ~(1 << BIT (best));
		result = into_notmsgset (result, tmp);
		freemsgset (tmp);
	}

	result = into_andmsgset (result, domain);

	return (result);
}

/* ARGSUSED */
static struct msgset *wordmulti (index2, groups, fulldomain, domain,
	arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *arg;
int arglen;
int not;
int *perr;
{
	int meta;
	struct msgset *conc;
	struct msgset *result;
	char *p;
	int brace = 0;

	for (meta = 0, p = arg; *p != '\0'; p++) {
		if (*p == '*' || *p == '?') {
			meta = 1;
		} else if (*p == '[') {
			meta = 1;
			brace = 1;
		} else if (isalnum (*p)) {
			*p = tolower (*p);
		} else if (brace && (*p == ']' || *p == '!' || *p == '-')) {
			if (*p == ']')
				brace = 0;
		} else {
			*perr = 1;
			(void) printf (
				"?word query contains non-alphanumeric\n");
			return (0);
		}
	}

	conc = concord (index2, (struct msgset *) 0, arg,
		0, meta?SHELLMATCH:EXACTMATCH);

	if (not) {
		result = notmsgset (domain, conc);
		freemsgset (conc);
	} else
		result = into_andmsgset (conc, domain);

	return (result);
}

/* ARGSUSED */
static struct msgset *wordmentions (index2, groups, fulldomain, domain,
	arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *arg;
int arglen;
int not;
int *perr;
{
	int meta;
	struct msgset *conc;
	struct msgset *result;
	char *p;
	int brace = 0;

	for (meta = 0, p = arg; *p != '\0'; p++) {
		if (*p == '*' || *p == '?') {
			meta = 1;
		} else if (*p == '[') {
			meta = 1;
			brace = 1;
		} else if (isalnum (*p)) {
			*p = tolower (*p);
		} else if (brace && (*p == ']' || *p == '!' || *p == '-')) {
			if (*p == ']')
				brace = 0;
		} else {
			*perr = 1;
			(void) printf (
				"?word query contains non-alphanumeric\n");
			return (0);
		}
	}
	
	conc = concord (index2, (struct msgset *) 0, arg,
		0, meta?SHELLMATCH:EXACTMATCH);

	if (not) {
		result = notmsgset (domain, conc);
		freemsgset (conc);
	} else
		result = into_andmsgset (conc, domain);

	return (result);
}

#define MAXRANGE 10

/* ARGSUSED */
static struct msgset *pathnamefunc (index2, groups, fulldomain, domain,
	arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *arg;
int arglen;
int not;
int *perr;
{
	struct msgset *result = 0, *tmp;
	register char *p, *q;
	struct messagepat mpat;
	register struct message *m;
	register long i, n;
	register long x;
	register long msg;
	long maxmsg;
	long *bitset;
	char group[100];
	char *slash;
	static char dir[] = "/usr/spool/news/";
	int l, c;

	if (arglen > sizeof (group) - 1)
		return (0);

	mpat.message = 0;
	mpat.grouplist = malloc ((unsigned) groups->ngroup);
	if (mpat.grouplist == 0) {
		perror ("indexserver: pathnamefunc malloc");
		exit (2);
	}
	bzero (mpat.grouplist, groups->ngroup);

	for (p = arg, q = group, slash = 0; *p != '\0'; p++) {
		if (*p == '.' || *p == '/') {
			if (slash != q) {
				*q++ = '/';
				slash = q;
			}
		} else
			*q++ = *p;
	}
	*q = '\0';
	p = slash;
	if (p == 0)
		p = group;
	for (q = p; *q != '\0' && isdigit (*q); q++)
		;
	if (*q == '\0' && isdigit (*p)) {
		mpat.message = atol (p);
		if (p > group && p[-1] == '/')
			p--;
		*p = '\0';
	}
	if (strncmp (group, dir, strlen (dir)) == 0)
		(void) strcpy (group, group+strlen (dir));

	if (group[0] != '\0') {
		p = groups->text.text.data;
		l = strlen (group);
		for (i = 0; i < groups->ngroup; i++) {
			if (strncmp (group, p + groups->group[i], l) == 0 &&
			    ((c = *(p+groups->group[i]+l)) == 0 || c == '/'))
				mpat.grouplist[i] = 1;
		}
	} else if (mpat.message == 0) {
		free (mpat.grouplist);
		return (0);
	} else {
		for (i = 0; i < groups->ngroup; i++)
			mpat.grouplist[i] = 1;
	}

	if (countmsgset (domain) < 2000) {
		result = selectmsg (index2, domain, msgselect,
			(char *) &mpat, not);
		free (mpat.grouplist);
		return (result);
	}

	maxmsg = 0;
	bitset = 0;
	msg = n = index2->indices[0].summary.messages.ndata /
		sizeof (struct message);
	m = (struct message *) index2->indices[0].summary.messages.data;

	if (mpat.message == 0) {
		p = mpat.grouplist;
		for (i = 0; i < n; i++, m++) {
			if (p[m->group]) {
				if (i >= maxmsg) {
					result = extendmsgset (result,
						(int)ROUNDUP (i+200000));
					maxmsg = result->nbitset * LONGBITS;
					bitset = result->bitset;
				}
				bitset[ROUNDDOWN (i)] |= (1 << BIT (i));
			}
		}
	} else {
		x = mpat.message;
		p = mpat.grouplist;
		for (i = 0; i < n; i++, m++) {
			if (x == m->message && p[m->group]) {
				if (i >= maxmsg) {
					result = extendmsgset (result,
						(int)ROUNDUP (i+200000));
					maxmsg = result->nbitset * LONGBITS;
					bitset = result->bitset;
				}
				bitset[ROUNDDOWN (i)] |= (1 << BIT (i));
			}
		}
	}

	n = index2->indices[1].summary.messages.ndata / sizeof (struct message);
	m = (struct message *) index2->indices[1].summary.messages.data;
	x = mpat.message;
	p = mpat.grouplist;
	for (i = 0; i < n; i++, m++) {
		if ((x == 0 || x == m->message) && p[m->group]) {
			if (i+msg >= maxmsg) {
				result = extendmsgset (result,
					(int)ROUNDUP (i+msg+200000));
				maxmsg = result->nbitset * LONGBITS;
				bitset = result->bitset;
			}
			bitset[ROUNDDOWN (i+msg)] |= (1 << BIT (i+msg));
		}
	}

	free (mpat.grouplist);

        if (not) {
                tmp = notmsgset (domain, result);
                freemsgset (result);
                result = tmp;
        } else
                result = into_andmsgset (result, domain);

	return (result);
}

/* ARGSUSED */
static struct msgset *onfunc (index2, groups, fulldomain, domain,
	arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *arg;
int arglen;
int not;
int *perr;
{
	int day, month, year;
	char datebuf[20];
	struct msgset *result = 0;
	char str[20];

	if (parsedate (arg, &day, &month, &year, str)) {
		if (month < 0)
			(void) sprintf (datebuf, "%03x", year);
		else if (day < 0)
			(void) sprintf (datebuf, "%03x%1x", year, month);
		else
			(void) sprintf (datebuf, "%03x%1x%02x", year,month,day);

		result = dateset (index2, domain, datebuf, 0, not);
	}

	return (result);
}

/* ARGSUSED */
static struct msgset *afterfunc (index2, groups, fulldomain, domain,
	arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *arg;
int arglen;
int not;
int *perr;
{
	int day, month, year;
	char datebuf[20];
	struct msgset *result = 0;
	char str[20];
	
	if (parsedate (arg, &day, &month, &year, str)) {
		if (month < 0)
			month = 12, day = 31;
		else if (day < 0)
			day = 31;
		(void) sprintf (datebuf, "%03x%1x%02x", year, month, day);
		result = dateset (index2, domain, datebuf, 1, not);
	}

	return (result);
}

/* ARGSUSED */
static struct msgset *beforefunc (index2, groups, fulldomain, domain,
	arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *arg;
int arglen;
int not;
int *perr;
{
	int day, month, year;
	char datebuf[20];
	struct msgset *result = 0;
	char str[20];
	
	if (parsedate (arg, &day, &month, &year, str)) {
		if (month < 0)
			month = 0, day = 0;
		else if (day < 0)
			day = 0;
		(void) sprintf (datebuf, "%03x%1x%02x", year, month, day);
		result = dateset (index2, domain, datebuf, -1, not);
	}

	return (result);
}
